// D3DInitialize.cpp: implementation of the CD3DInitialize class.
//
//////////////////////////////////////////////////////////////////////

#include "D3DInitialize.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CD3DInitialize::CD3DInitialize()
{
	Cursor = false;

	num_modes   = 0;
	num_formats = 0;  

	////////////////////////////////////////////
	//hardware capabilities
	max_texture_stages = 0;
	max_texture_width = 0;
	max_texture_height = 0;
	max_active_lights = 0;
	max_streams = 0;
	max_point_size = 0;

	windowed_acceleration = false;
	pure_device = false;
	hardware_TNL = false;
	hardware_rasterization = false;
	does_texgen = false;
	does_zbias = false;
	stencil_support = false;	
}

CD3DInitialize::~CD3DInitialize()
{

}


bool CD3DInitialize::InitializeD3D(HWND hwnd, int w, int h, bool fullscreen, LPDIRECT3DDEVICE8 &device, LPDIRECT3D8 &d3d){

	HRESULT rv = 0;
	int j;

	int width  = w;
	int height = h;

	D3DDISPLAYMODE desktop_mode;
	D3DFORMAT DepthStencilFormat = D3DFMT_UNKNOWN;
	D3DFORMAT BackBufferFormat   = D3DFMT_UNKNOWN;
	D3DFORMAT DisplayFormat      = D3DFMT_UNKNOWN;
	UINT AdapterType             = D3DADAPTER_DEFAULT;
	D3DDEVTYPE DeviceType        = D3DDEVTYPE_HAL;

	bool found		 = false;

	int bit_depth    = 16;
	int alpha_bits   = 0;
	int stencil_bits = 8;
	int depth_bits   = 24;

	RECT rect;
	RECT client_rect;

	///////////////////////////////////////////////
	//get basic window info
    GetWindowRect(hwnd, &rect);
    GetClientRect(hwnd, &client_rect);

	///////////////////////////////////////////////
	//initialize Direct3D
	d3d = Direct3DCreate8(D3D_SDK_VERSION);
	if(d3d == NULL)
		return false;


	///////////////////////////////////////////////
	//get adapter display mode
	if(!fullscreen)
	{
		//get the current desktop display mode
		d3d->GetAdapterDisplayMode(AdapterType, &desktop_mode);
		DisplayFormat    = desktop_mode.Format;
		BackBufferFormat = desktop_mode.Format;

		BuildDisplayModeList(d3d);
	}
	else
	{
		//build the list of supported display modes
		num_modes = BuildDisplayModeList(d3d);

		//does the adapter support the requested display dimensions?
		found = false;
		for(j = 0; j < num_modes; j++)
		{
			if(width == DisplayModes[j].Width && height == DisplayModes[j].Height)
			{
				found = true;
				break;
			}
		}
		if(!found)
		{
			LogError("Adapter does not support the required dimensions to run this app");
			return false;
		}

		/////////////////////////////////////////////////////////////////////
		//pick a display format

		//prefer a 32 bit display format if possible
		D3DFORMAT df_preferred   = D3DFMT_X8R8G8B8;
		D3DFORMAT df_secondary   = D3DFMT_X8R8G8B8;
		found = false;
		for(j = 0; j < num_formats; j++)
		{
			if((D3DFMT_X8R8G8B8 == Formats[j]) || (D3DFMT_A8R8G8B8 == Formats[j]))
			{
				DisplayFormat = D3DFMT_X8R8G8B8;
				found         = true;
				break;
			}
		}

		//if cannot do 32 bit support, try a 16 bit format
		if(!found)
		{
			LogError("Adapter does not support a 32bit display mode. Trying 16bit.");

			if(alpha_bits > 0)
			{
				df_preferred = D3DFMT_X4R4G4B4;
				df_secondary = D3DFMT_X1R5G5B5;
			}
			else
			{
				df_preferred = D3DFMT_R5G6B5;
				df_secondary = D3DFMT_R5G6B5;
			}

			//first try to find a match for our preferred display format
			found = false;
			for(j = 0; j < num_formats; j++)
			{
				if(Formats[j] == df_preferred)
				{
					DisplayFormat = df_preferred;
					found = true;
					break;
				}
			}

			//if preference not found, try our second choice display format
			if(!found)
			{
				for(j = 0; j < num_formats; j++)
				{
					if(Formats[j] == df_secondary)
					{
						DisplayFormat = df_secondary;
						found = true;
						break;
					}
				}
			}
		}

		//if we cannot find a suitable fullscreen format, abort initialization
		if(!found)
		{
			LogError("Could not find a suitable fullscreen format");
			return false;
		}
	}

	if((DisplayFormat == D3DFMT_X8R8G8B8) || (DisplayFormat == D3DFMT_A8R8G8B8))
		bit_depth = 32;

	/////////////////////////////////////////////////////////////////
	//find a suitable backbuffer format for this display format
	if(fullscreen)
	{
		if(!GetBackBufferFormat(bit_depth, alpha_bits, DisplayFormat, &BackBufferFormat))
		{
			LogError("Search could not find an appropriate backbuffer format."); 
		}
		if(BackBufferFormat == D3DFMT_UNKNOWN)
		{
			LogError("No suitable backbuffer format found. Exiting.");
			return false;
		}

		//is hardware acceleration supported for the selected display/backbuffer format?
		rv = d3d->CheckDeviceType(AdapterType,
								  DeviceType,
 								  DisplayFormat,
  								  BackBufferFormat,
								  FALSE);
		if(rv != D3D_OK)
		{
			LogError("Acceleration not supported on display/backbuffer format");
			return false;
		}
	}

	////////////////////////////////////////////////////////////
	//select an appropriate stencil/depth buffer surface format
	DepthStencilFormat = GetDepthBufferFormat(d3d, DisplayFormat, false, bit_depth);
	if(DepthStencilFormat == D3DFMT_UNKNOWN)
	{
		LogError("No suitable depth buffer found. Exiting.");
		return false;
	}

	///////////////////////////////////////////////////////////
	//initialize our presentation parameters
	D3DPRESENT_PARAMETERS d3d_pp;
	memset(&d3d_pp,0,sizeof(D3DPRESENT_PARAMETERS));
	d3d_pp.BackBufferWidth  = width;
	d3d_pp.BackBufferHeight = height;
	d3d_pp.BackBufferFormat = BackBufferFormat;
	d3d_pp.BackBufferCount  = 1;
	d3d_pp.MultiSampleType  = D3DMULTISAMPLE_NONE;
	d3d_pp.SwapEffect       = D3DSWAPEFFECT_DISCARD;//D3DSWAPEFFECT_DISCARD;D3DSWAPEFFECT_FLIP
	d3d_pp.hDeviceWindow    = hwnd;
	d3d_pp.Windowed         = fullscreen ? FALSE : TRUE;
	d3d_pp.EnableAutoDepthStencil = TRUE;
	d3d_pp.AutoDepthStencilFormat = DepthStencilFormat;
	d3d_pp.Flags = 0;
	d3d_pp.FullScreen_RefreshRateInHz      = 0;
	d3d_pp.FullScreen_PresentationInterval = 0;

	////////////////////////////////////////////////////////////
	//create the device (we dont use a pure device)
	DWORD device_behavior[4] = 
	{
		D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE,
		D3DCREATE_HARDWARE_VERTEXPROCESSING,
		D3DCREATE_MIXED_VERTEXPROCESSING,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING
	};
	for(j = 1; j < 4 ; j++)
	{
		rv = d3d->CreateDevice(AdapterType,
							   DeviceType,
							   hwnd, 
							   device_behavior[j],
							   &d3d_pp, 
							   &device);
		if(rv == D3D_OK)
			break; 
		else
			device = NULL;
	}
	if(!device)
	{
		LogError("Could not create a suitable device. Exiting.");
		return false;
	}
	else
	{
		switch(j)
		{
			case 0:
				LogInfo("Pure device created");
			break;
			case 1:
				LogInfo("Hardware device created");
			break;
			case 2:
				LogInfo("Mixed HW/SW device created");
			break;
			case 3:
				LogInfo("Software device created");
			break;
		}
	}

	//////////////////////////////////////////////////////////
	//create the viewport
	D3DVIEWPORT8 d3dviewport;
	memset(&d3dviewport,0,sizeof(D3DVIEWPORT8));
	d3dviewport.X = 0;
	d3dviewport.Y = 0;
	d3dviewport.Width  = width;
	d3dviewport.Height = height;
	d3dviewport.MinZ   = 0.0F;
	d3dviewport.MaxZ   = 1.0F;
	if(FAILED(device->SetViewport(&d3dviewport)))
	{
		LogError("Failed to set the device viewport. Exiting.");
		return false;
	}

	///////////////////////////////////////////////////////////
	//set the device cursor
	if(fullscreen)
	{
		ShowCursor(Cursor);
		device->ShowCursor(Cursor);
	}
	else
	{
        SetWindowPos(hwnd, 
				     HWND_NOTOPMOST, 
                     rect.left,
					 rect.top,
                     rect.right  - rect.left, 
                     rect.bottom - rect.top, 
					 SWP_SHOWWINDOW );
	}

	//////////////////////////////////////////////////////
	//check device capabilities
	EvaluateDeviceCapabilities(device);

	return true;
}

int CD3DInitialize::BuildDisplayModeList(LPDIRECT3D8 d3d)
{
	//get the fullscreen display modes available on adapter 
	int j;
	bool found = false;

	num_modes = d3d->GetAdapterModeCount(D3DADAPTER_DEFAULT);
	for(int mode_num = 0; mode_num < num_modes; mode_num++)
	{
		//get the display mode
		D3DDISPLAYMODE dmode;
		d3d->EnumAdapterModes(D3DADAPTER_DEFAULT, mode_num, &dmode);

		//filter out lower resolutions
		if(dmode.Height < 480 || dmode.Width  < 640)
			continue;

		//filter out refresh rate redundancies for dimensions and format
		found = false;
		for(j = 0; j < DisplayModes.size(); j++){
			if(dmode.Width == DisplayModes[j].Width && dmode.Height == DisplayModes[j].Height){
				found = true;
				break;
			}
		}
		if(!found)
			DisplayModes.push_back(dmode);

		found = false;
		for(j = 0; j < num_formats; j++)
		{
			if(dmode.Format == Formats[j])
			{
				found = true;
				break;
			}
		}
		if(!found)
		{
			Formats[num_formats] = dmode.Format;
			num_formats++;
		}
	}
	num_modes = DisplayModes.size();

	for(j = 0; j < num_modes; j++)
	{
		LogInfo("%d x %d @%d",DisplayModes[j].Width, DisplayModes[j].Height, DisplayModes[j].Format);
	}

	return num_modes;
}

bool CD3DInitialize::GetBackBufferFormat(int bit_depth, int alpha_bits, D3DFORMAT display_format, D3DFORMAT *bb_format)
{
	if(bit_depth == 16)
	{
		LogError("Using 16 bit backbuffer");

		if(display_format == D3DFMT_X1R5G5B5)
		{
			if(alpha_bits > 0)
				*bb_format = D3DFMT_A1R5G5B5;
			else
				*bb_format = D3DFMT_X1R5G5B5;

			return true;
		}
		if(display_format == D3DFMT_X4R4G4B4)
		{
			if(alpha_bits > 0)
				*bb_format = D3DFMT_A4R4G4B4;
			else
				*bb_format = D3DFMT_X4R4G4B4;

			return true;
		}
		if(display_format == D3DFMT_R5G6B5)
		{
			*bb_format = D3DFMT_R5G6B5;
			if(alpha_bits > 0)
				return false;
			else
				return true;
		}
	}
	else if(bit_depth == 32)
	{

		LogError("Using 32 bit backbuffer");

		if(display_format == D3DFMT_X8R8G8B8)
		{
			if(alpha_bits > 0)
				*bb_format = D3DFMT_A8R8G8B8;
			else
				*bb_format = D3DFMT_X8R8G8B8;

			return true;
		}
		if(display_format == D3DFMT_R8G8B8)
		{
			*bb_format = D3DFMT_R8G8B8;
			if(alpha_bits > 0)
				return false;
			else
				return true;
		}
	}

	*bb_format = D3DFMT_UNKNOWN;
	return false;
}

D3DFORMAT CD3DInitialize::GetDepthBufferFormat(LPDIRECT3D8 d3d, D3DFORMAT DisplayFormat, bool use_stencil, int bpp)
{
	//if we are in 32bpp mode we will try to create a 32bit depth stencil surface
	//the fallback is a 16bit depth buffer (which is also required in 16bpp display mode on many devices)

	UINT AdapterType       = D3DADAPTER_DEFAULT;
	D3DDEVTYPE DeviceType  = D3DDEVTYPE_HAL;
	HRESULT hr = 0;

	if(bpp == 32)
	{
		hr = d3d->CheckDeviceFormat(AdapterType,
									DeviceType,
									DisplayFormat,
									D3DUSAGE_DEPTHSTENCIL,
									D3DRTYPE_SURFACE,
									D3DFMT_D24S8);
		if(FAILED(hr))
		{
			LogError("Could not create D24S8 depth buffer...trying D24X4S4...");
		}
		else return D3DFMT_D24S8;

		hr = d3d->CheckDeviceFormat(AdapterType,
									DeviceType,
									DisplayFormat,
									D3DUSAGE_DEPTHSTENCIL,
									D3DRTYPE_SURFACE,
									D3DFMT_D24X4S4);
		if(FAILED(hr))
		{
			LogError("Could not create D24X4S4 depth buffer...trying D24X8...");
		}
		else return D3DFMT_D24X4S4;

		hr = d3d->CheckDeviceFormat(AdapterType,
									DeviceType,
									DisplayFormat,
									D3DUSAGE_DEPTHSTENCIL,
									D3DRTYPE_SURFACE,
									D3DFMT_D24X8);
		if(FAILED(hr))
		{
			LogError("Could not create D24X8 depth buffer...trying D32...");
		}
		else return D3DFMT_D24X8;


		hr = d3d->CheckDeviceFormat(AdapterType,
									DeviceType,
									DisplayFormat,
									D3DUSAGE_DEPTHSTENCIL,
									D3DRTYPE_SURFACE,
									D3DFMT_D32);
		if(FAILED(hr))
		{
			LogError("Could not create D32 depth buffer...trying D16...");
		}
		else return D3DFMT_D32;
	}

	LogInfo("Using 16bit depth buffer");
	hr = d3d->CheckDeviceFormat(AdapterType,
								DeviceType,
								DisplayFormat,
								D3DUSAGE_DEPTHSTENCIL,
								D3DRTYPE_SURFACE,
								D3DFMT_D16);
	if(FAILED(hr))
	{
		LogError("Could not create D16 depth buffer... exiting application.");
	}
	else return D3DFMT_D16;

	return D3DFMT_UNKNOWN;
}


bool CD3DInitialize::EvaluateDeviceCapabilities(const LPDIRECT3DDEVICE8 device)
{
	//some useful information to keep handy
	int   max_texture_stages = 0;
	int   max_texture_width = 0;
	int   max_texture_height = 0;
	int   max_active_lights = 0;
	int   max_streams = 0;
	float max_point_size = 0;

	bool windowed_acceleration = false;
	bool pure_device = false;
	bool hardware_TNL = false;
	bool hardware_rasterization = false;
	bool does_texgen = false;
	bool does_zbias = false;
	bool stencil_support = false;

	/////////////////////////////////////////////////
	//retrieve device capabilities
	D3DCAPS8 device_caps;
	device->GetDeviceCaps(&device_caps);

	/////////////////////////////////////////////////
	//check hardware support for our various features
	if(device_caps.DevCaps & D3DDEVCAPS_HWRASTERIZATION)
		hardware_rasterization = true;

	if(device_caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT)
		hardware_TNL = true;

	if(device_caps.DevCaps & D3DDEVCAPS_PUREDEVICE)
		pure_device = true;

	if(device_caps.Caps2 & D3DCAPS2_CANRENDERWINDOWED)
		windowed_acceleration = true;

	if(device_caps.VertexProcessingCaps & D3DVTXPCAPS_TEXGEN)
		does_texgen = true;

	if(device_caps.RasterCaps & D3DPRASTERCAPS_ZBIAS)
		does_zbias = true;

	LPDIRECT3DSURFACE8 dbuffer = NULL;
	device->GetDepthStencilSurface(&dbuffer);
	if(dbuffer)
	{
		D3DSURFACE_DESC desc;
		dbuffer->GetDesc(&desc);
		if(desc.Format == D3DFMT_D24S8)
			stencil_support = true;
	}

	SAFE_RELEASE(dbuffer);

	max_texture_stages = device_caps.MaxSimultaneousTextures;
	max_active_lights  = device_caps.MaxActiveLights;
	max_point_size     = device_caps.MaxPointSize;
	max_streams        = device_caps.MaxStreams;
	max_texture_width  = device_caps.MaxTextureWidth;
	max_texture_height = device_caps.MaxTextureHeight;

	//are there any engine settings we need to adjust based on these caps? (test here)
	
	return true;
	
}


